from django.conf import settings
from rest_framework import serializers


class HyperlinkedModelSerializer(serializers.HyperlinkedModelSerializer):
    @staticmethod
    def fix_view_name(field_kwargs, model):
        if 'view_name' in field_kwargs:
            view_name_suffix = field_kwargs['view_name'].split('-')[-1]
            field_kwargs['view_name'] = "{}:{}-{}".format(settings.API_APP_NAMESPACE,
                                                          model._meta.label_lower,
                                                          view_name_suffix)
        return field_kwargs

    def build_url_field(self, field_name, model_class):
        field_class, field_kwargs = super().build_url_field(field_name, model_class)
        return field_class, self.fix_view_name(field_kwargs, self.Meta.model)

    def build_relational_field(self, field_name, relation_info):
        field_class, field_kwargs = super().build_relational_field(field_name, relation_info)
        return field_class, self.fix_view_name(field_kwargs, model=relation_info[1])



from rest_framework import routers


class RestRouter(routers.DefaultRouter):
    def get_default_base_name(self, viewset):
        """
        If `base_name` is not specified, attempt to automatically determine
        it from the viewset.
        """
        queryset = getattr(viewset, 'queryset', None)

        assert queryset is not None, '`base_name` argument not specified, and could ' \
            'not automatically determine the name from the viewset, as ' \
            'it does not have a `.queryset` attribute.'

        return queryset.model._meta.label_lower



import json

from asgiref.sync import async_to_sync
from channels.generic.websocket import AsyncWebsocketConsumer
from channels.layers import get_channel_layer
from django.urls import NoReverseMatch
from rest_framework.reverse import reverse

from channels_redux.AbstractNotifier import AbstractNotifier

ats = async_to_sync


class NotifyConsumer(AsyncWebsocketConsumer):
    @property
    def user(self):
        return self.scope["user"]

    async def connect(self):
        if self.user.is_anonymous:
            await self.close()

        await self.accept()

    async def receive(self, text_data=None, bytes_data=None):
        json_data = json.loads(text_data)
        operation = {
            "subscribe.new": self.subscribe_new,
            "subscribe.existing": self.subscribe_existing,
        }.get(json_data["type"], None)

        if operation is None:
            raise NotImplementedError("Operation {} is not supported. Received data {}".format(operation, json_data))
        await operation(json_data)

    async def subscribe_new(self, message: dict):
        await self.subscribe_to_group(message['model'])

    async def subscribe_existing(self, message: dict):
        model = message['model']
        ids = message['ids']
        for pk in ids:
            await self.subscribe_to_group(self.group_name(model, pk))

    async def subscribe_to_group(self, group_name):
        await self.channel_layer.group_add(group_name, self.channel_name)

    async def notify(self, event):
        await self.send(text_data=json.dumps({
            "model": event['model'],
            "pk": event['pk'],
            "type": event['subtype'],
            "url": event['url']
        }))

    @staticmethod
    def group_name(model, pk=None):
        return '{}.{}'.format(model, pk) if pk else model

    @staticmethod
    def object_updated(instance: AbstractNotifier):
        channel_layer = get_channel_layer()
        message = NotifyConsumer.message_for_type("updated", instance)
        ats(channel_layer.group_send)(NotifyConsumer.group_name(message["model"], instance.pk), message)

    @staticmethod
    def object_created(instance: AbstractNotifier):
        channel_layer = get_channel_layer()
        message = NotifyConsumer.message_for_type("created", instance)
        ats(channel_layer.group_send)(NotifyConsumer.group_name(message["model"]), message)

    @staticmethod
    def object_deleted(instance: AbstractNotifier):
        channel_layer = get_channel_layer()
        message = NotifyConsumer.message_for_type("deleted", instance)
        ats(channel_layer.group_send)(NotifyConsumer.group_name(message["model"], instance.pk), message)

    @staticmethod
    def message_for_type(message_type, instance: AbstractNotifier):
        try:
            return {
                "type": "notify",
                "subtype": message_type,
                "model": instance._meta.label_lower,
                "pk": instance.pk,
                "url": reverse('{}-detail'.format(instance.get_api_base_name()), args=(instance.pk,)),
            }
        except NoReverseMatch:
            print('API url does not exist for {}'.format(instance.get_api_base_name()))



from django.conf import settings
from django.db import models


class AbstractNotifier(models.Model):
    class Meta:
        abstract = True

    @classmethod
    def get_api_base_name(cls):
        return '{}:{}'.format(settings.API_APP_NAMESPACE, cls._meta.label_lower)



from channels.routing import URLRouter
from django.conf.urls import url

from channels_redux import NotifyConsumer


class NotifyRouter(URLRouter):
    def __init__(self, path=r'^ws/notify/$', name='ws-notify'):
        super().__init__([url(path, NotifyConsumer, name=name)])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="channels-redux",
    description="Channels-Redux is a package that notifies subscribers about changes in the database",
    version="1.1",
    license='MIT License',
    include_package_data=True,
    author="Johnathan Ryan Hornik",
    author_email="ryanhornik@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/plus1tables/channels-redux-python/",
    packages=setuptools.find_packages(),
    classifiers=(
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ),
    install_requires=[
        "Django>=1.11",
        "djangorestframework>=3.4.0",
        "channels>=2.0.0",
    ]
)



from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver

from channels_redux.NotifyConsumer import NotifyConsumer
from channels_redux.AbstractNotifier import AbstractNotifier
from channels_redux.NotifyRouter import NotifyRouter

__all__ = [
    'NotifyConsumer',
    'AbstractNotifier',
    'NotifyRouter'
]


@receiver(post_save)
def object_saved(sender, instance: AbstractNotifier, created, **kwargs):
    if not issubclass(sender, AbstractNotifier):
        return
    if created:
        NotifyConsumer.object_created(instance)
    else:
        NotifyConsumer.object_updated(instance)


@receiver(post_delete)
def object_deleted(sender, instance: AbstractNotifier, **kwargs):
    if not issubclass(sender, AbstractNotifier):
        return
    NotifyConsumer.object_deleted(instance)
